use warnings;
use strict;

package Business::XPDL;
use base 'XML::Compile::Cache';
our $VERSION = '0.01';

use XML::Compile::Util qw/type_of_node unpack_type/;
use Log::Report 'business-xpdl', syntax => 'SHORT';

=head1 NAME

Business::XPDL - XPDL access

=head1 SYNOPSIS

  my $xpdl = Business::XPDL->new(version => '2.1');
  $xpdl->importDefinitions(...);
  my $xml  = $xpdl->writer($sometype)->($doc, $perldata);

  my ($type, $data) = $xpdl->from('data.xml');
  my ($type, $data) = Business::XPDL->from('data.xml');
  print "Root type: $type\n";
  print Dumper $data;    # from Data::Dumper

=head1 DESCRIPTION

This module provides access to the Process Description Language, as
defined by the WfMC, the Workflow Management Coalition (F<wfmc.org>)

=head2 Limitiations

The XPDL files may contain additional external type definitions, which
are not imported/included in the specification.  You need to import
those explictly by hand.  See F<t/20sample.t>

TypeDeclarations are part of the interpretation of the messages which are
exchanged between the Applications, and therefore not interpreted by
this module.

=cut

use Business::XPDL::Util;

# map namespace always to the newest implementation of the protocol
my %ns2version =
 ( &NS_XPDL_009 => '0.09'
 , &NS_XPDL_10  => '1.0'
 , &NS_XPDL_20  => '2.0'
 , &NS_XPDL_21  => '2.1'
 );

my %info =
 ( '0.01' => { }  # not usable
 , '0.09' =>
     { prefixes => { xpdl => NS_XPDL_009 }
     }
 , '1.0' =>
     { prefixes => { xpdl => NS_XPDL_10 }
     }
 , '2.0alpha-21' =>
     { prefixes => { xpdl => NS_XPDL_20 }
     }
 , '2.0alpha-24' =>
     { prefixes => { xpdl => NS_XPDL_20 }
     }
 , '2.0'         =>   # alpha namespace used for final product
     { prefixes => { xpdl => NS_XPDL_20 }
     }
 , '2.1' =>
     { prefixes => { xpdl => NS_XPDL_21 }
     }
 );

=head1 Constructors

=over 4

=item new OPTIONS

Boolean OPTION C<support_deprecated> will invoke the type declarations
of older XPDL standard versions as well.  It is allowed to use those in
XPDL messages, but not supported by (probably) many design tools.

Boolean OPTION C<prefix_keys> enables the visibility of the various
XPDL namespaces which can be merged within one document.
=cut

sub new($)
{   my $class = shift;
    $class->SUPER::new(direction => 'RW', @_);
}

sub init($)
{   my ($self, $args) = @_;
    $args->{allow_undeclared} = 1
        unless exists $args->{allow_undeclared};

    $args->{opts_rw}   = { @{$args->{opts_rw}} }
        if ref $args->{opts_rw} eq 'ARRAY';
    $args->{opts_rw}{mixed_elements}  = 'STRUCTURAL';
    $args->{opts_rw}{sloppy_floats}   = 1;
    $args->{opts_rw}{sloppy_integers} = 1;
    $args->{any_element}            ||= 'ATTEMPT';

    $self->SUPER::init($args);

    my $version = $args->{version}
        or error __x"XPDL object requires an explicit version";

    unless(exists $info{$version})
    {   exists $ns2version{$version}
            or error __x"XPDL version {v} not recognized", v => $version;
        $version = $ns2version{$version};
    }
    $self->{version} = $version;
    my $info = $info{$version};

    my $prefix_keys = delete $args->{prefix_keys};

    $self->prefixes(%{$info->{prefixes}});
    $self->addKeyRewrite('PREFIXES(xpdl)') if $prefix_keys;

    (my $xsd = __FILE__) =~ s!\.pm!/xsd!;
    my @xsds = glob "$xsd/xpdl-$version/*";

    if(delete $args->{support_deprecated})
    {   if($version gt '1.0')   # $version is a version label, not number
        {   $self->prefixes(xpdl10 => NS_XPDL_10);
            push @xsds, glob "$xsd/xpdl-1.0/*";
            $self->addKeyRewrite('PREFIXES(xpdl10)') if $prefix_keys;
        }
        if($version ge '2.1')
        {   $self->prefixes(xpdl20 => NS_XPDL_20);
            push @xsds, glob "$xsd/xpdl-2.0alpha-final/*";
            $self->addKeyRewrite('PREFIXES(xpdl20)') if $prefix_keys;
        }
    }

    $self->importDefinitions(\@xsds);
    $self;
}

=item from XMLDATA, OPTIONS

Read an XML message, in any format supported by L<XML::Compile> method
C<dataToXML()>: string, file, filehandle, and more.  Returned is a list
of two: the type of the top-level element plus the data-structure.

When called as instance method, the data will automatically get converted
to the version of required by the object.  When called as class method,
the version of the top-level element will determine the returned version
automatically (which may give unpredictable versions as result).

When the method is called as class method, then a temporary instance is
created.  Creating an instance is (very) slow.

  my $xpdl          = Business::XPDL->new(version => '2.1');
  my ($type, $data) = $xpdl->from('data.xml');

or

  my ($type, $data) = Business::XPDL->from('data.xml');

=cut

sub from($@)
{   my ($thing, $data, %args) = @_;

    my $xml  = XML::Compile->dataToXML($data);
    my $top  = type_of_node $xml;
    my $ns   = (unpack_type $top)[0];
    my $version = $ns2version{$ns}
       or error __x"unknown XPDL version with namespace {ns}", ns => $ns;

    my $self;
    if(ref $thing)
    {   # instance method
        $self = $thing;
    }
    else
    {   # class method: can determine version myself
        $self = $thing->new(version => $version, %args);
    }

    my $r    = $self->reader($top, %args)
        or error __x"root node `{top}' not recognized", top => $top;

    ($top, $r->($xml));
}

=back

=head2 Accessors

=over 4

=item version

=back

=cut

sub version() {shift->{version}}

1;
